from pathlib import Path
from typing import TYPE_CHECKING

import napari
import numpy as np

from argclz import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOptions
from neuralib.util.verbose import fprint
from .core import read_stardist, StarDistSegmentation, STARDIST_MODEL

if TYPE_CHECKING:
    from stardist.models import StarDist2D

__all__ = ['StarDist2DOptions']


class StarDist2DOptions(AbstractSegmentationOptions):
    DESCRIPTION = 'Run the Stardist model for segmentation'

    model: STARDIST_MODEL = as_argument(AbstractSegmentationOptions.model).with_options(
        default='2D_versatile_fluo',
        help='stardist pretrained model'
    )

    prob_thresh: float | None = argument(
        '--prob',
        default=None,
        help='Consider only object candidates from pixels with predicted object probability above this threshold. '
             'Seealso: stardist.models.base._predict_instances_generator: prob_thresh'
    )

    def run(self):
        if self.napari_view:
            self.launch_napari()
        else:
            self.eval()

    def seg_output(self, filepath: Path) -> Path:
        return filepath.with_name(filepath.stem + '_seg').with_suffix('.npz')

    def eval(self, **kwargs) -> None:
        from stardist.models import StarDist2D

        model = StarDist2D.from_pretrained(self.model)

        if self.file_mode:
            self._eval(self.file, self.process_image(), model)
        elif self.batch_mode:
            for file, image in self.foreach_process_image():
                self._eval(file, image, model)
        else:
            raise RuntimeError('run fail')

    def _eval(self, filepath: Path, image: np.ndarray, model: 'StarDist2D', **kwargs):
        out_seg = self.seg_output(filepath)

        if out_seg.exists() and not self.invalid_existed_result:
            fprint(f'cached {filepath} because {out_seg} exists, use --invalid to invalid cache', vtype='IO')
            return

        labels, detail = model.predict_instances(image, prob_thresh=self.prob_thresh, **kwargs)
        labels = labels.astype(np.float_)
        labels[labels == 0] = np.nan

        res = StarDistSegmentation(labels, detail['coord'], detail['prob'], str(filepath), self.model)

        # mask probability
        if self.prob_thresh is not None:
            res.mask_probability(self.prob_thresh)

        # save output
        res.to_npz(out_seg)

        if self.save_ij_roi:
            res.to_roi(self.ij_roi_output(filepath))

    # noinspection PyTypeChecker
    def launch_napari(self, with_widget: bool = False):
        """
        Launch napari viewer for stardist results

        :param with_widget: If True, launch also with the starDist widget (required package ``stardist-napari``)
        """
        file = self.seg_output(self.file)
        if not file.exists() or self.invalid_existed_result:
            self.eval()

        res = read_stardist(file)

        viewer = napari.Viewer()
        viewer.add_image(self.process_image(), name='image')
        viewer.add_image(res.labels, name='labels', colormap='twilight_shifted', opacity=0.5)
        viewer.add_points(res.points, face_color='red')

        if with_widget:
            viewer.window.add_plugin_dock_widget("stardist-napari", "StarDist")

        napari.run()


if __name__ == '__main__':
    StarDist2DOptions().main()



from __future__ import annotations

import pickle
from pathlib import Path
from typing import TypedDict

import polars as pl

from neuralib.typing import PathLike
from neuralib.util.dataframe import DataFrameWrapper

__all__ = [
    'Joint',
    'read_dlc',
    'DeepLabCutDataFrame',
    'JointDataFrame',
    'DeepLabCutMeta',
    'DeepLabCutModelConfig'
]

Joint = str
"""Joint name"""


def read_dlc(file: PathLike, meta_file: PathLike | None = None) -> DeepLabCutDataFrame:
    """
    load DeepLabCut result from file

    :param file: DeepLabCut result filepath. supports both ``.h5`` and ``.csv``
    :param meta_file: Optional DeepLabCut meta filepath. should be the ``.pickle``
    :return:
    """
    file = Path(file)
    meta_file = Path(meta_file)
    meta = _load_meta(meta_file)

    match file.suffix:
        case '.h5' | '.hdf5':
            df = _load_dlc_h5_table(file)
        case '.csv':
            df = _load_dlc_csv(file)
        case _:
            raise ValueError(f'file: {file} is not supported')

    return DeepLabCutDataFrame(df, meta=meta, filtered=('filtered' in file.name))


class DeepLabCutDataFrame(DataFrameWrapper):
    """
    DeepLabCut DataFrame ::

        ┌───────────┬───────────┬───────────┬───────────┬───┬───────────┬───────────┬───────────┬──────────┐
        │ Nose_x    ┆ Nose_y    ┆ Nose_like ┆ EarL_x    ┆ … ┆ TailMid_l ┆ TailEnd_x ┆ TailEnd_y ┆ TailEnd_ │
        │ ---       ┆ ---       ┆ lihood    ┆ ---       ┆   ┆ ikelihood ┆ ---       ┆ ---       ┆ likeliho │
        │ f64       ┆ f64       ┆ ---       ┆ f64       ┆   ┆ ---       ┆ f64       ┆ f64       ┆ od       │
        │           ┆           ┆ f64       ┆           ┆   ┆ f64       ┆           ┆           ┆ ---      │
        │           ┆           ┆           ┆           ┆   ┆           ┆           ┆           ┆ f64      │
        ╞═══════════╪═══════════╪═══════════╪═══════════╪═══╪═══════════╪═══════════╪═══════════╪══════════╡
        │ 57.907318 ┆ 512.54742 ┆ 0.999679  ┆ 77.701355 ┆ … ┆ 0.999904  ┆ 257.71426 ┆ 561.89660 ┆ 0.999961 │
        │           ┆ 4         ┆           ┆           ┆   ┆           ┆ 4         ┆ 6         ┆          │
        │ 57.907318 ┆ 516.79528 ┆ 0.999688  ┆ 77.701355 ┆ … ┆ 0.999923  ┆ 257.71426 ┆ 562.05725 ┆ 0.999954 │
        │ …         ┆ …         ┆ …         ┆ …         ┆ … ┆ …         ┆ …         ┆ …         ┆ …        │
        │ 94.259621 ┆ 43.849434 ┆ 0.973851  ┆ 106.33532 ┆ … ┆ 0.998977  ┆ 87.477776 ┆ 257.11996 ┆ 0.999937 │
        │ 94.294357 ┆ 44.340511 ┆ 0.965436  ┆ 106.45220 ┆ … ┆ 0.999604  ┆ 87.223534 ┆ 258.46600 ┆ 0.999912 │
        └───────────┴───────────┴───────────┴───────────┴───┴───────────┴───────────┴───────────┴──────────┘

    """

    def __init__(self, df: pl.DataFrame,
                 meta: DeepLabCutMeta | None, *,
                 filtered: bool):
        """
        :param df: DeepLabCut result dataframe
        :param meta: :attr:`~DeepLabCutMeta`
        :param filtered: whether the results has already been filtered
        """
        self._df = df
        self._meta = meta
        self._filtered = filtered

    def __repr__(self):
        raise repr(self.dataframe())

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        if dataframe is None:
            return self._df
        else:
            return DeepLabCutDataFrame(dataframe, meta=self._meta, filtered=self._filtered)

    @property
    def default_filtered(self) -> bool:
        """whether default filtered when running the deeplabcut"""
        return self._filtered

    @property
    def meta(self) -> DeepLabCutMeta:
        """:attr:`~neuralib.tracking.deeplabcut.DeepLabCutMeta`"""
        return self._meta

    @property
    def model_config(self) -> DeepLabCutModelConfig:
        """:attr:`~neuralib.tracking.deeplabcut.DeepLabCutModelConfig`"""
        return self.meta['model_config']

    @property
    def fps(self) -> float:
        """frame per second, meta data required"""
        return self.meta['fps']

    @property
    def nframes(self) -> int:
        """number of frames"""
        return self.meta['nframes']

    @property
    def joints(self) -> list[Joint]:
        """list of labelled joints"""
        return self.meta['model_config']['all_joints_names']

    def get_joint(self, joint: Joint) -> JointDataFrame:
        """get specific joint"""
        cols = ('x', 'y', 'likelihood')
        df = self.select([f'{joint}_{col}' for col in cols]).dataframe()
        return JointDataFrame(df)


class JointDataFrame(DataFrameWrapper):
    """
    Dataframe from a specific joint ::

        ┌───────────┬────────────┬─────────────────┐
        │ Nose_x    ┆ Nose_y     ┆ Nose_likelihood │
        │ ---       ┆ ---        ┆ ---             │
        │ f64       ┆ f64        ┆ f64             │
        ╞═══════════╪════════════╪═════════════════╡
        │ 57.907318 ┆ 512.547424 ┆ 0.999679        │
        │ 57.907318 ┆ 516.795288 ┆ 0.999688        │
        │ 57.907318 ┆ 519.56311  ┆ 0.999449        │
        │ 56.733799 ┆ 522.204224 ┆ 0.999161        │
        │ 53.546089 ┆ 525.24939  ┆ 0.999518        │
        │ …         ┆ …          ┆ …               │
        │ 94.259621 ┆ 43.849434  ┆ 0.973851        │
        │ 94.294357 ┆ 44.111595  ┆ 0.980125        │
        │ 94.8013   ┆ 44.340511  ┆ 0.963981        │
        │ 94.294357 ┆ 44.340511  ┆ 0.947905        │
        │ 94.294357 ┆ 44.340511  ┆ 0.965436        │
        └───────────┴────────────┴─────────────────┘
    """

    def __init__(self, df: pl.DataFrame):
        self._df = df

    def __repr__(self):
        return repr(self.dataframe())

    def dataframe(self, dataframe: pl.DataFrame = None, may_inplace=True):
        if dataframe is None:
            return self._df
        else:
            return JointDataFrame(dataframe)


class DeepLabCutMeta(TypedDict):
    """DeepLabCut model metadata"""
    start: float
    stop: float
    run_duration: float
    Scorer: str
    model_config: DeepLabCutModelConfig
    fps: float
    batch_size: int
    frame_dimensions: tuple[int, int]
    nframes: int
    iteration: int
    training_set_fraction: float
    cropping: bool
    cropping_parameters: list[tuple[float, float, float, float]]


class DeepLabCutModelConfig(TypedDict):
    """DeepLabCut model configuration"""
    stride: float
    weigh_part_predictions: bool
    weigh_negatives: bool
    fg_fraction: float
    mean_pixel: list[float]
    shuffle: bool
    snapshot_prefix: str
    log_dir: str
    global_scale: float
    location_refinement: bool
    locref_stdev: float
    locref_loss_weight: float
    locref_huber_loss: bool
    optimizer: str
    intermediate_supervision: bool
    intermediate_supervision_layer: int
    regularize: bool
    weight_decay: float
    crop_pad: int
    scoremap_dir: str
    batch_size: int
    dataset_type: str
    deterministic: bool
    mirror: bool
    pairwise_huber_loss: bool
    weigh_only_present_joints: bool
    partaffinityfield_predict: bool
    pairwise_predict: bool
    all_joints: list[list[int]]
    all_joints_names: list[Joint]
    dataset: str
    init_weights: str
    net_type: str
    num_joints: int
    num_outputs: int


def _load_dlc_h5_table(file) -> pl.DataFrame:
    import pandas as pd

    df = pd.read_hdf(file)

    scorers = list(df.columns.levels[0])
    bodyparts = list(df.columns.levels[1])
    coords = list(df.columns.levels[2])

    assert len(scorers) == 1
    scorer = scorers[0]

    data = {
        f'{b}_{c}': df[(scorer, b, c)]
        for b in bodyparts
        for c in coords
    }

    ret = pl.DataFrame(data)

    return ret


def _load_dlc_csv(file) -> pl.DataFrame:
    cols = ['']
    with file.open() as f:
        f.readline()  # skip first line
        parts = f.readline().strip().split(',')[1::3]
        cols.extend([f'{p}_{it}' for p in parts for it in ('x', 'y', 'likelihood')])

    df = pl.read_csv(file, skip_rows=3, has_header=False, new_columns=cols)[:, 1:]

    return df


def _load_meta(meta_file) -> DeepLabCutMeta:
    if meta_file.suffix not in ('.pkl', '.pickle'):
        raise ValueError(f'{meta_file} is not a pickle file')

    # meta
    with meta_file.open('rb') as f:
        meta = pickle.load(f)['data']

    # copy to typeddict
    meta['model_config'] = meta['DLC-model-config file']
    meta['iteration'] = meta['iteration (active-learning)']
    meta['training_set_fraction'] = meta['training set fraction']

    return meta



from pathlib import Path

import numpy as np
import polars as pl
from brainglobe_atlasapi import BrainGlobeAtlas
from brainrender.actors import Points
from scipy.interpolate import interp1d
from typing_extensions import Self

from argclz import argument
from neuralib.atlas.brainrender.core import BrainRenderCLI
from neuralib.atlas.typing import PLANE_TYPE
from neuralib.atlas.util import allen_to_brainrender_coord
from neuralib.typing import PathLike

__all__ = ['ProbeRenderCLI',
           'ProbeShank']


class ProbeRenderCLI(BrainRenderCLI):
    """Probe track reconstruction with brainrender"""
    DESCRIPTION = 'Probe track reconstruction with brainrender'

    GROUP_PROBE = 'Probe Option'

    implant_depth: int = argument(
        '--depth',
        group=GROUP_PROBE,
        help='implant depth in um'
    )

    shank_interval: int | None = argument(
        '--interval',
        default=None,
        group=GROUP_PROBE,
        help='shank interval in um if multi-shank'
    )

    dye_label_only: bool = argument(
        '--dye',
        group=GROUP_PROBE,
        help='only show the histology dye parts'
    )

    remove_outside_brain: bool = argument(
        '--remove-outside-brain',
        group=GROUP_PROBE,
        help='remove reconstruction outside the brain'
    )

    file: Path = argument(
        '--file',
        metavar='FILE',
        type=Path,
        required=True,
        group=GROUP_PROBE,
        help='multi-shank npy or csv file to be inferred'
    )

    plane_type: PLANE_TYPE = argument(
        '--plane-type', '-P',
        default='coronal',
        group=GROUP_PROBE,
        help='cutting orientation to infer the multi-shank label point/probe_idx'
    )

    def post_parsing(self):
        super().post_parsing()
        if not self.dye_label_only and self.implant_depth is None:
            raise ValueError('')

    def run(self):
        if not self._stop_render:
            self.render()
            self.render_output()

    def render(self):
        self.post_parsing()
        super().render()
        self._add_probe()

    def _add_probe(self):
        bg = self.get_atlas_brain_globe()

        if self.file.suffix == '.csv':
            probe = ProbeShank.load_csv(self.file, self.plane_type, bg)
        elif self.file.suffix == '.npy':
            probe = ProbeShank.load_numpy(self.file, bg)
        else:
            raise TypeError(f"Unsupported file type: {self.file.suffix}")

        #
        if self.coordinate_space == 'ccf':
            probe = probe.map_brainrender()

        #
        if not self.dye_label_only:
            probe_theo = probe.as_theoretical(self.implant_depth, self.shank_interval, self.remove_outside_brain)
            self.scene.add(Points(probe_theo, colors='k', name='theo', alpha=0.9))

        probe_dye = probe.interp(ret_type=np.ndarray)
        self.scene.add(Points(probe_dye, colors='r', name='dye', alpha=0.9))


class ProbeShank:
    """shank reconstruction class"""

    def __init__(self, dorsal: np.ndarray,
                 ventral: np.ndarray,
                 bg: BrainGlobeAtlas):
        """

        :param dorsal: `Array[float, 3 | [S, 3]]`
        :param ventral: `Array[float, 3 | [S, 3]]`
        :param bg: ``BrainGlobeAtlas``
        """
        self._dorsal = dorsal
        self._ventral = ventral
        self._validate()

        self._bg = bg

    def _validate(self):
        assert self._dorsal.shape == self._ventral.shape

    def __iter__(self):
        """foreach shank DV"""
        for i in range(self.n_shanks):
            yield self._dorsal[i], self._ventral[i]

    @classmethod
    def load_numpy(cls, file: PathLike, bg: BrainGlobeAtlas) -> Self:
        """Load numpy array. `Array[float, [2, 3] | [S, 2, 3]]`

        ``S`` = Number of shanks. If 2D then single shank
        ``2`` = Dorsal and ventral
        ``3`` = AP, DV, ML coordinates
        """
        data = np.load(file)
        if data.ndim == 2:
            return cls(data[0], data[1], bg)
        elif data.ndim == 3:
            return cls(data[:, 0, :], data[:, 1, :], bg)
        else:
            raise ValueError(f'not support {data.shape=}')

    @classmethod
    def load_csv(cls, file: PathLike,
                 plane_type: PLANE_TYPE,
                 bg: BrainGlobeAtlas,
                 verbose: bool = True) -> Self:
        """
        Load from csv file

        :param file: csv file
        :param plane_type: {'coronal', 'sagittal', 'transverse'}
        :param bg: BrainGlobeAtlas
        :param verbose:
        :return:
        """
        df = pl.read_csv(file)
        cols = ['AP_location', 'DV_location', 'ML_location']
        if 'probe_idx' not in df.columns or 'point' not in df.columns:
            df = df.select(cols)
            df = cls._sort_rows(df, plane_type)

        if verbose:
            from neuralib.util.verbose import printdf
            printdf(df)

        data = df.select(cols).to_numpy().reshape(-1, 2, 3)
        return cls(data[:, 0, :], data[:, 1, :], bg)

    @classmethod
    def _sort_rows(cls, df: pl.DataFrame, plane_type: PLANE_TYPE) -> pl.DataFrame:
        n = int(df.shape[0] / 2)
        expr = pl.Series(['dorsal'] * n + ['ventral'] * n)
        df = df.sort('DV_location').with_columns(expr.alias('point'))

        if plane_type == 'sagittal':
            shank_order = 'AP_location'
        elif plane_type == 'coronal':
            shank_order = 'ML_location'
        else:
            raise ValueError('')

        df = (df.sort(by=['point', shank_order], descending=[False, True])
              .with_columns(pl.Series(list(range(1, 1 + n)) * 2).alias('probe_idx'))
              .sort('probe_idx'))

        return df

    @property
    def n_shanks(self) -> int:
        """number of shanks"""
        return self._dorsal.shape[0]

    def map_brainrender(self) -> Self:
        """map allen ccf brain space to brainrender"""
        self._dorsal = allen_to_brainrender_coord(self._dorsal)
        self._ventral = allen_to_brainrender_coord(self._ventral)
        return self

    def interp(self, interp_range: tuple[float, float] | None = None,
               ret_type: type = np.ndarray) -> list[np.ndarray] | np.ndarray:
        """extend_depth foreach shank

        :param interp_range:
        :param ret_type: if as list, then list[Array[float, [P, 3]]]. if numpy array `Array[float, [P * S, 3]]`
        :return: list[Array[float, [P, 3]]] | `Array[float, [P * S, 3]]`
        """
        ret = [self._interp(d, v, interp_range) for d, v in self]

        if ret_type is list:
            return ret
        elif ret_type is np.ndarray:
            return np.vstack(ret)
        else:
            raise ValueError('')

    def as_theoretical(self, depth: int, interval: int | None = None, remove_outside_brain: bool = True) -> np.ndarray:
        """
        as theoretical array

        :param depth: implanted depth
        :param interval: interval between shanks
        :param remove_outside_brain: remove the point outside the brain
        :return:
        """
        shanks = self.interp(interp_range=(0, 5000), ret_type=list)

        ret = []
        for shank in shanks:
            p = self._within_depth_range(shank, depth, remove_outside_brain)
            ret.append(p)

            if interval is not None:
                depth += self._calc_shank_length_diff(shanks[0], interval)
                interval += interval

        return np.vstack(ret)

    def _interp(self, d, v, interp_range: tuple[float, float] | None):
        """

        :param d: `Array[float, 3]`
        :param v: `Array[float, 3]`
        :param interp_range: interpolation DV in um. usually a large value, then cut afterward
        :return: `Array[float, [P, 3]]`
        """
        if interp_range is not None:
            nn = np.arange(*interp_range, 10)
        else:
            nn = np.arange(d[1], v[1], 10)

        s = np.vstack([d, v])

        return interp1d(s[:, 1], s, axis=0, bounds_error=False, fill_value='extrapolate')(nn)

    def _within_depth_range(self, shank: np.ndarray,
                            depth: int,
                            remove_outside_brain: bool = True):
        """remove point segments outside the brain, and more than given depth"""

        shank = shank[shank[:, 1] >= 0]

        if remove_outside_brain:
            mx = self._isin_brain(shank)
            shank = shank[mx]

        d = np.sqrt(np.sum((shank - shank[0]) ** 2, axis=1))

        return shank[d <= depth]

    def _isin_brain(self, shank: np.ndarray) -> np.ndarray:
        """

        :param shank: `Array[float, [P, 3]]`
        :param bg: ``BrainGlobeAtlas``
        :return: `Array[bool, P]`
        """
        bg = self._bg

        ret = []
        for p in shank:
            try:
                rid = bg.annotation[bg._idx_from_coords(p, microns=True)]
            except IndexError:
                rid = 0
            ret.append(rid != 0)

        return np.array(ret, dtype=bool)

    @staticmethod
    def _calc_shank_length_diff(shank: np.ndarray, shank_interval: float) -> float:
        """
        use the vector of the probe, then calculate the unit vector

        :param shank: `Array[float, [P, 3]]`
        :param shank_interval: distance (um) relative to the specific shank. e.g., NeuroPixel 2.0 = 250 * x
        :return: unit vector
        """
        v = shank[-1] - shank[0]
        v = v / np.linalg.norm(v)  # unit vector

        # vector product: |n x v| = sin_theta |n| |v|
        # inline n value, got following formula
        sin_theta = np.linalg.norm(np.array([v[2], 0, -v[0]]))

        return shank_interval * sin_theta


if __name__ == '__main__':
    ProbeRenderCLI().main()



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.colors import Normalize

from neuralib.plot import plot_figure
from neuralib.plot.colormap import get_customized_cmap
from neuralib.tracking.facemap import FaceMapResult
from neuralib.typing import ArrayLikeStr, PathLike

__all__ = ['plot_facemap_keypoints',
           'plot_cmap_time_series']


def plot_facemap_keypoints(fmap: FaceMapResult,
                           frame_interval: tuple[int, int],
                           keypoints: str | ArrayLikeStr | None = None,
                           outlier_filter: bool = True,
                           output: PathLike | None = None) -> None:
    """
    Plot all the keypoints

    :param fmap: :class:`~neuralib.wrapper.facemap.core.FaceMapResult`
    :param frame_interval: frame interval (start/end) for the plot
    :param keypoints: a keypoint name or multiple keypoints. If None, then show all the keypoints
    :param outlier_filter: remove jump and do the interpolation
    :param output: output file. Show the fig if None
    """
    if keypoints is None:
        kps = fmap.keypoints
    else:
        kps = list(keypoints)

    mx = np.arange(*frame_interval)
    colors = get_customized_cmap('jet', (0, 1), len(kps))

    with plot_figure(output) as ax:
        df = fmap.get(*kps)
        if outlier_filter:
            df = df.with_outlier_filter()

        for i, it in enumerate(df.partition_by('keypoint')):
            name = it['keypoint'].unique().item()
            ax.plot(mx, it['x'][mx], '-', color=colors[i], label=f'{name}_x')
            ax.plot(mx, it['y'][mx], '--', color=colors[i], label=f'{name}_y')

        ax.set(xlabel='frame', ylabel='keypoint coordinates')
        ax.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.)


def plot_cmap_time_series(x: np.ndarray,
                          y: np.ndarray, *,
                          cmap: str = 'viridis',
                          ax: Axes | None = None,
                          with_color_bar: bool = True,
                          color_bar_label: str = 'frames',
                          output: PathLike | None = None,
                          **kwargs) -> None:
    """
    Plots a scatter plot with a colorbar

    :param x: X-axis values. `Array[float, T]`
    :param y: Y-axis values. `Array[float, T]`
    :param cmap: Colormap to use
    :param ax: ``matplotlib.axes.Axes``
    :param with_color_bar: if show the colorbar
    :param color_bar_label: color bar label
    :param output: Output file path to save the plot
    :param kwargs: pass through ``ax.scatter()``
    """
    if x.size != y.size:
        raise ValueError('size xy inconsistent')

    colors = get_customized_cmap(cmap, value=(0, 1), numbers=len(x))
    norm = Normalize(vmin=0, vmax=len(x))

    def _plot(ax: Axes):
        ax.scatter(x, y, c=colors, alpha=0.5, **kwargs)
        if with_color_bar:
            cbar = plt.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=ax)
            cbar.set_label(color_bar_label)
            ax.set(xlabel='x', ylabel='y')

    #
    if ax is None:
        with plot_figure(output) as ax:
            _plot(ax)
    else:
        _plot(ax)
        if output is not None:
            plt.savefig(output)



import warnings

warnings.warn(
    "The 'sqlp' module is deprecated and will be removed in a separated module. "
    "Use 'pip install sqlclz' instead.'",
    DeprecationWarning,
    stacklevel=2
)

from . import util
from .annotation import *
from .cli import *
from .connection import *
from .func import *
from .func_date import *
from .func_stat import *
from .func_win import *
from .literal import *
from .stat import Cursor
from .stat_start import *
from .table import *
from .table_nt import *



from __future__ import annotations

from pathlib import Path
from typing import NamedTuple, Any, get_args

import attrs
import cv2
import imageio.v3 as iio
import numpy as np
import polars as pl
from scipy.io import loadmat
from scipy.io.matlab import MatlabOpaque

from neuralib.atlas.typing import PLANE_TYPE
from neuralib.atlas.view import SlicePlane, get_slice_view
from neuralib.imglib.transform import apply_transformation
from neuralib.typing import PathLike

__all__ = [
    'SLICE_DIMENSION_10um',
    'CCFTransMatrix',
    'load_transform_matrix',
    'slice_transform_helper'
]

SLICE_DIMENSION_10um: dict[PLANE_TYPE, tuple[int, int]] = {
    'coronal': (1140, 800),
    'sagittal': (1320, 800)
}


def load_transform_matrix(filepath: PathLike,
                          plane_type: PLANE_TYPE,
                          resolution: int = 10,
                          default_name: str = 'test') -> CCFTransMatrix:
    """
    matrix for image transformation

    :param filepath: transformation matrix .mat file
    :param plane_type: ``PLANE_TYPE`` {'coronal', 'sagittal', 'transverse'}
    :param resolution: atlas resolution
    :param default_name: name of the ``CCFTransMatrix``
    :return: ``CCFTransMatrix``
    """
    mat = loadmat(filepath, squeeze_me=True, struct_as_record=False)['save_transform']
    f = Path(filepath).name

    try:
        name = f[:f.index('_resize')]
    except ValueError:
        name = default_name

    return CCFTransMatrix(
        name,
        MatMatrix(mat.allen_location, mat.transform, mat.transform_points),
        plane_type,
        resolution=resolution
    )


class MatMatrix(NamedTuple):
    allen_location: np.ndarray
    """abnormal parsing in loadmat. [996 array([  0, -12], dtype=int16)]"""
    transform: MatlabOpaque
    """mat meta info. currently not used"""
    transform_points: np.ndarray
    """[(P, 2), (P, 2)] transformation point for 2dccf GUI site view"""

    @property
    def slice_index(self) -> int:
        return int(self.allen_location[0])

    # noinspection PyTypeChecker
    @property
    def delta_values(self) -> tuple[int, int]:
        """shifting values for yx axis (height, width) for a specific plane type.

        Coronal slice: [0]:DV, [1]:ML

        Sagittal Slice: [0]:DV, [1]:AP

        """
        delta = self.allen_location[1]
        return delta[0], delta[1]


@attrs.frozen
class CCFTransMatrix:
    """matrix for image transformation"""

    slice_id: str
    """slice name id"""
    matrix: MatMatrix
    """MatMatrix"""
    plane_type: PLANE_TYPE
    """PLANE_TYPE"""
    resolution: int = attrs.field(kw_only=True, default=10)
    """resolution in um"""

    @property
    def slice_index(self) -> int:
        """slice index"""
        return self.matrix.slice_index

    @property
    def delta_xy(self) -> tuple[int, int]:
        """map delta value to xy slice view"""
        return int(self.matrix.delta_values[1]), int(self.matrix.delta_values[0])

    def get_slice_plane(self) -> SlicePlane:
        """get slice plane"""
        dw = self.delta_xy[0]
        dh = self.delta_xy[1]

        ret = (
            get_slice_view('reference', self.plane_type, resolution=self.resolution)
            .plane_at(self.slice_index)
            .with_offset(
                dw=dw + 1 if dw != 0 else 0,  # avoid index err
                dh=dh + 1 if dh != 0 else 0,  # avoid index err
            )
        )

        return ret

    def matrix_info(self, to_polars: bool = True) -> dict[str, Any] | pl.DataFrame:
        """
        transform matrix information dict

        :param to_polars: to polars dataframe
        :return:
        """
        ret = {
            'slice_id': self.slice_id,
            'plane_type': self.plane_type,
            'atlas_resolution': self.resolution,
            'slice_index': self.slice_index,
            'reference_value': self.get_slice_plane().reference_value,
            'dw': self.delta_xy[0],
            'dh': self.delta_xy[1]
        }

        if to_polars:
            ret = pl.DataFrame(ret)

        return ret


def slice_transform_helper(raw_image: PathLike | np.ndarray,
                           trans_matrix: np.ndarray | PathLike, *,
                           plane_type: PLANE_TYPE = 'coronal',
                           flip_lr: bool = False,
                           flip_ud: bool = False) -> tuple[np.ndarray, np.ndarray]:
    """
    Transforms an input image according to the specified transformation matrix,
    plane orientation, and optional flipping parameters. This function reads a raw
    image, optionally flips it horizontally or vertically, applies a transformation
    matrix, and resizes it to a defined dimension based on the plane type. The
    result is a tuple containing the raw image and the transformed image.

    :param raw_image: Path to the raw input image or image array.
    :param trans_matrix: Transformation matrix to apply to the image. It can be a
        numpy array or the path to a valid file containing the matrix. Supported
        files are `.mat` for MATLAB files and `.npy` for NumPy files.
    :param plane_type: Defines the anatomical plane for transformation.
        Defaults to 'coronal'.
    :param flip_lr: Boolean flag to determine whether to flip the image
        horizontally (left to right).
    :param flip_ud: Boolean flag to determine whether to flip the image
        vertically (up to down).
    :return: A tuple containing two numpy arrays: the raw image as processed,
        and the transformed image.
    """
    if isinstance(raw_image, np.ndarray):
        pass
    elif isinstance(raw_image, get_args(PathLike)):
        raw_image = iio.imread(raw_image)
    else:
        raise TypeError(f"Unsupported type {type(raw_image)}")

    #
    if flip_ud:
        raw_image = np.flipud(raw_image)

    if flip_lr:
        raw_image = np.fliplr(raw_image)

    #
    if isinstance(trans_matrix, get_args(PathLike)):
        s = Path(trans_matrix).suffix
        if s == '.mat':
            mtx = loadmat(trans_matrix)['t'].T
        elif s == '.npy':
            mtx = np.load(trans_matrix)
        else:
            raise ValueError(f'invalid file type for the transformation matrix: {s}')
    else:
        mtx = trans_matrix

    resize = cv2.resize(raw_image, SLICE_DIMENSION_10um[plane_type])
    transform_image = apply_transformation(resize, mtx)

    return raw_image, transform_image



from __future__ import annotations

import functools
import sqlite3
import warnings
from collections.abc import Iterator
from typing import overload, TYPE_CHECKING, Any, TypeVar, Generic, Optional, Literal, Union, cast, Callable

import polars as pl
from typing_extensions import Self

from .expr import *
from .expr import SqlStatBuilder, SqlRemovePlaceHolder
from .table import *

if TYPE_CHECKING:
    from .connection import Connection

__all__ = [
    'SqlStat',
    'SqlSelectStat',
    'SqlInsertStat',
    'SqlUpdateStat',
    'SqlDeleteStat',
    'Cursor',
]

T = TypeVar('T')
S = TypeVar('S')


def catch_error(f=None, *, attr: str = None):
    def _catch_error_decorator(f):
        @functools.wraps(f)
        def _catch_error(self, *args, **kwargs):
            if attr is None:
                stat: SqlStat = self
            else:
                stat = getattr(self, attr)

            with stat:
                return f(self, *args, **kwargs)

        return _catch_error

    if f is None:
        return _catch_error_decorator
    else:
        return _catch_error_decorator(f)


class SqlStat(Generic[T]):
    """Abstract SQL statement."""

    def __init__(self, table: Optional[type[T]]):
        self.table = table
        self._stat: list[Any] = []

        from .connection import get_connection_context
        self._connection: Optional[Connection] = get_connection_context()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            self.drop()

    def build(self) -> tuple[str, list]:
        """build a SQL statement."""
        builder = SqlStatBuilder()
        builder.add(self)
        self._stat = None
        return builder.build(), builder.para

    def __str__(self) -> str:
        table = table_name(self.table) if self.table is not None else '...'
        return f'{type(Self).__name__}[{table}]'

    def __repr__(self) -> str:
        builder = SqlStatBuilder()
        builder._deparameter = 'repr'
        builder.add(self)
        return builder.build()

    def submit(self) -> Cursor[T]:
        """
        build the SQL statement and execute.

        :return: a cursor
        :raise RuntimeError: current statement does not bind with a connection.
        """
        if (connection := self._connection) is None:
            raise RuntimeError('Do not in a connection context')

        ret = Cursor(connection, connection.execute(self), self.table)
        self._connection = None
        return ret

    def drop(self):
        self._connection = None
        self._stat = None

    def __del__(self):
        # check connection and auto_commit,
        # make SqlStat auto submit itself when nobody is referring to it,
        # so users do need to explict call submit() for every statements.
        # FIXME error raised during statement constructing should be aware
        if self._connection is not None and self._stat is not None and len(self._stat) > 0:
            try:
                self.submit()
            except BaseException as e:
                warnings.warn(repr(e))

    def add(self, stat) -> Self:
        """
        Add SQL token.
        """
        if self._stat is None:
            raise RuntimeError("Statement is closed.")

        self._stat.append(stat)
        if isinstance(stat, SqlStat):
            stat._connection = None

        return self


class Cursor(Generic[T]):
    """
    A SQL cursor wrapper.

    It will try to cast to T from the tuple returns.

    """

    def __init__(self, connection: Connection, cursor: sqlite3.Cursor, table: type[T] = None):
        self._connection = connection
        self._cursor = cursor

        if table is not None:
            from .table import table_class
            try:
                table_cls = table_class(table)
            except AttributeError:
                pass
            else:
                cursor.row_factory = lambda _, row: table_cls.table_new(*row)

    @property
    def headers(self) -> list[str]:
        header = self._cursor.description
        return [it[0] for it in header] if header is not None else []

    def __del__(self):
        self._cursor.close()
        self._connection = None

    def fetchall(self) -> list[T]:
        """fetch all results."""
        return list(self)

    def fetchone(self) -> Optional[T]:
        """fetch the first result."""
        if (ret := self._cursor.fetchone()) is None:
            return None

        return ret

    def __iter__(self) -> Iterator[T]:
        """iterate the results."""
        yield from iter(self._cursor)

    def fetch_polars(self) -> pl.DataFrame:
        return pl.DataFrame(list(self._cursor), schema=self.headers)


class SqlWhereStat:
    """statement with **WHERE** support."""

    @catch_error
    def where(self, *expr: Union[bool, SqlCompareOper, SqlExpr, None]) -> Self:
        """
        ``WHERE`` clause: https://www.sqlite.org/lang_select.html#whereclause

        >>> select_from(A).where( # doctest: SKIP
        ...     A.a == 1, A.b == 2
        ... ).build()
        SELECT * FROM A
        WHERE (A.a = 1) AND (A.b = 2)

        :param expr:
        :return:
        """
        zelf = cast(SqlStat, self)
        expr = [it for it in expr if it is not None]
        if len(expr):
            from .func_stat import and_
            zelf.add('WHERE')
            zelf.add(and_(*expr))
        return self


class SqlLimitStat:
    @overload
    def limit(self, n: int) -> Self:
        pass

    @overload
    def limit(self, row_count: int, offset: int) -> Self:
        pass

    @catch_error
    def limit(self, *args: int) -> Self:
        """
        ``LIMIT``: https://www.sqlite.org/lang_select.html#limitoffset

        >>> select_from(A).limit(10).build() # doctest: SKIP
        SELECT * FROM A LIMIT 10
        >>> select_from(A).limit(10, 10).build() # doctest: SKIP
        SELECT * FROM A LIMIT 10 OFFSET 10

        **NOTE**

        LIMIT on UPDATE/DELETE need compile flag ``SQLITE_ENABLE_UPDATE_DELETE_LIMIT``.

        >>> assert Connection().sqlite_compileoption_used('SQLITE_ENABLE_UPDATE_DELETE_LIMIT') # doctest: SKIP

        """
        zelf = cast(SqlStat, self)

        if len(args) == 1:
            n, *_ = args
            zelf.add(['LIMIT', str(n)])
        elif len(args) == 2:
            row_count, offset = args
            zelf.add(['LIMIT', str(row_count), 'OFFSET', str(offset)])
        else:
            raise TypeError()
        return self

    @catch_error
    def order_by(self, *by: Union[int, str, SqlExpr, Any]) -> Self:
        """
        ``ORDER BY``: https://www.sqlite.org/lang_select.html#orderby

        >>> select_from(A).order_by(A.a).build() # doctest: SKIP
        SELECT * FROM A ORDER BY A.

        **possible ordering**

        >>> select_from(A).order_by( # doctest: SKIP
        ...     asc(A.a), desc(A.b), nulls_first(A.c), asc(A.d).nulls_last(),
        ... )
        SELECT * FROM A ORDER BY
        A.a ASC, A.b DESC, A.c NULLS FIRST, A.d ASC NULLS LAST

        **NOTE**

        ORDER BY on UPDATE/DELETE need compile flag ``SQLITE_ENABLE_UPDATE_DELETE_LIMIT``.

        >>> assert Connection().sqlite_compileoption_used('SQLITE_ENABLE_UPDATE_DELETE_LIMIT') # doctest: SKIP

        """
        zelf = cast(SqlStat, self)
        zelf.add('ORDER BY')
        fields = []
        for it in by:
            if isinstance(it, int):
                fields.append(SqlLiteral(str(it)))
            elif isinstance(it, str):
                fields.append(SqlLiteral(it))
            elif isinstance(it, SqlAlias):
                fields.append(SqlLiteral(it._name))
            elif isinstance(it, SqlExpr):
                fields.append(it)
            else:
                fields.append(wrap(it))

        zelf.add(SqlConcatOper(fields, ','))

        return self


class SqlSelectStat(SqlStat[T], SqlWhereStat, SqlLimitStat, Generic[T]):
    """**SELECT** statement."""

    def __init__(self, table: Optional[type[T]]):
        super().__init__(table)
        self._involved: list[Union[type, SqlAlias[type]]] = []

    def __matmul__(self, other: str) -> SqlAlias[SqlSubQuery]:
        """wrap itself as a subquery with an alias name."""
        return SqlAlias(SqlSubQuery(self), other)

    def fetchall(self) -> list[T]:
        """submit and fetch all result."""
        return self.submit().fetchall()

    def fetchone(self) -> Optional[T]:
        """submit and fetch the first result."""
        return self.submit().fetchone()

    def __iter__(self) -> Iterator[T]:
        """submit and iterate the results"""
        return iter(self.submit())

    def fetch_polars(self) -> pl.DataFrame:
        """submit and fetch all as a polar dataframe."""
        return self.submit().fetch_polars()

    @catch_error
    def windows(self, *windows: Union[SqlWindowDef, SqlAlias[SqlWindowDef]], **window_ks: SqlWindowDef) -> Self:
        """
        define windows.

        """
        if len(windows) == 0 and len(window_ks) == 0:
            return self

        self.add('WINDOW')

        for window in windows:
            if isinstance(window, SqlWindowDef):
                if window.name is None:
                    raise RuntimeError('?? AS ' + repr(window))
                self.add([window.name, 'AS'])
                self.add(window)
            elif isinstance(window, SqlAlias) and isinstance(window._value, SqlWindowDef):
                self.add([window.name, 'AS'])
                self.add(window)
            else:
                raise TypeError()

            self.add(',')

        for name, window in window_ks.items():
            self.add([name, 'AS'])
            self.add(window)
            self.add(',')
        self._stat.pop()

        return self

    BY = Literal['left', 'right', 'inner', 'full outer', 'cross']

    @overload
    def join(self, constraint: Union[Callable, ForeignConstraint], *, by: BY = None) -> SqlSelectStat[tuple]:
        pass

    @overload
    def join(self, table: Union[type[S], SqlAlias[S]],
             constraint: Union[Callable, ForeignConstraint], *,
             by: BY = None) -> SqlSelectStat[tuple]:
        pass

    @overload
    def join(self, table: Union[type[S], SqlSelectStat[S], SqlAlias[S], SqlCteExpr],
             *field: Union[bool, Any],
             by: BY = None) -> SqlSelectStat[tuple]:
        pass

    @overload
    def join(self, *field: bool | Any, by: BY = None) -> SqlSelectStat[tuple]:
        pass

    @catch_error
    def join(self, *args, by: BY = None) -> SqlSelectStat[tuple]:
        """
        ``JOIN`` https://www.sqlite.org/lang_select.html#strange_join_names

        >>> select_from(A.a, B.b).join(A.a == B.a) # doctest: SKIP
        SELECT A.a, B.b FROM A
        JOIN B ON A.a = B.a

        """
        if by is not None:
            self.add(by.upper())
        self.add('JOIN')
        self.table = None

        if len(args) == 2 and isinstance(table := args[0], type) and (
                callable(constraint := args[1]) or isinstance(constraint, ForeignConstraint)):
            if not isinstance(constraint, ForeignConstraint):
                if (constraint := table_foreign_field(constraint)) is None:
                    raise RuntimeError('not a foreign constraint')
            self.__join_foreign(constraint, table)

        elif len(args) > 0 and isinstance(table := args[0], type):
            self.__join(table, *args[1:])

        elif len(args) > 0 and isinstance(expr := args[0], SqlCteExpr):
            self._stat.insert(0, expr)
            self.__join(expr, *args[1:])

        elif len(args) > 0 and isinstance(stat := args[0], SqlSelectStat):
            self.add(stat)
            self.__join(None, *args[1:])

        elif len(args) == 2 and isinstance(table := args[0], SqlAlias) and (
                callable(constraint := args[1]) or isinstance(constraint, ForeignConstraint)):
            if not isinstance(constraint, ForeignConstraint):
                if (constraint := table_foreign_field(constraint)) is None:
                    raise RuntimeError('not a foreign constraint')
            self.__join_foreign(constraint, table)

        elif len(args) > 0 and isinstance(table := args[0], SqlAlias) and isinstance(table._value, type) and isinstance(
                table._name, str):
            self.__join(table, *args[1:])

        elif len(args) > 0 and isinstance(table := args[0], SqlAlias) and isinstance(expr := table._value,
                                                                                     SqlCteExpr) and isinstance(table._name, str):
            self._stat.insert(0, expr)
            self.__join(table, *args[1:])

        elif len(args) > 0 and isinstance(table := args[0], SqlAlias) \
                and isinstance(subq := table._value, SqlSubQuery) \
                and isinstance(name := table._name, str) \
                and isinstance(stat := subq.stat, SqlSelectStat):
            self.add(stat)
            self.add('AS')
            self.add(name)
            self.__join(None, *args[1:])

        elif len(args) == 1 and isinstance(constraint := args[0], ForeignConstraint):
            self.__join_foreign(constraint, None)

        elif len(args) == 1 and callable(constraint := args[0]):
            if (constraint := table_foreign_field(constraint)) is None:
                raise RuntimeError('not a foreign constraint')
            self.__join_foreign(constraint, None)

        else:
            table = None
            for field in args:
                if table is None and isinstance(field, SqlExpr):
                    table = use_table(field, self._involved)
            if table is None:
                raise RuntimeError('no join table')
            self.__join(table, *args)

        return self

    def __join(self, right: Union[type, SqlAlias, SqlCteExpr, None], *fields):
        if right is None:
            pass
        elif isinstance(right, type):
            self.add(table_name(right))
        elif isinstance(right, SqlCteExpr):
            self.add(right._name)
        elif isinstance(right, SqlAlias) and isinstance(table := right._value, type):
            self.add(table_name(table))
            self.add(right._name)
        elif isinstance(right, SqlAlias) and isinstance(expr := right._value, SqlCteExpr):
            self.add(expr._select)
            self.add(right._name)
        else:
            raise TypeError(f'JOIN {right}')

        if len(fields):
            if any([isinstance(it, SqlCompareOper) for it in fields]):
                self.__join_on(*fields)
            else:
                self.__join_use(*fields)

        self._involved.append(right)

    def __join_use(self, *fields):
        self.add(['USING', '('])

        for field in fields:
            if isinstance(field, str):
                self.add(repr(field))
            if isinstance(field, Field):
                self.add(field.name)
            if isinstance(field, SqlField):
                self.add(field.name)
            else:
                raise TypeError('USING ' + repr(field))
            self.add(',')

        self._stat.pop()
        self.add(')')

    def __join_on(self, *exprs: bool | SqlExpr):
        self.add('ON')
        from .func_stat import and_
        self.add(and_(*exprs))

    def __join_foreign(self, constraint: ForeignConstraint, right: type | SqlAlias | None):
        this = that = None

        for _this in self._involved:
            if isinstance(_this, type) and _this == constraint.table:
                this = constraint.table_name
                if right is None:
                    right = constraint.foreign_table
                break
            elif isinstance(_this, type) and _this == constraint.foreign_table:
                that = constraint.table_name
                if right is None:
                    right = constraint.table
                break
            elif isinstance(_this, SqlAlias) and isinstance(table := _this._value, type) and table == constraint.table:
                this = _this._name
                if right is None:
                    right = constraint.foreign_table
                break
            elif isinstance(_this, SqlAlias) and isinstance(table := _this._value, type) and table == constraint.foreign_table:
                that = _this._name
                if right is None:
                    right = constraint.table
                break
        else:
            raise RuntimeError('improper foreign constraint')

        if isinstance(right, type):
            self.add(table_name(right))
            if this is None and right == constraint.table:
                this = _this.__name__
            if that is None and right == constraint.foreign_table:
                that = _this.__name__
        elif isinstance(right, SqlAlias) and isinstance(table := right._value, type):
            self.add(table_name(table))
            self.add(right._name)

            if this is None and right == constraint.table:
                this = right._name
            if that is None and right == constraint.foreign_table:
                that = right._name
        else:
            raise TypeError(f'JOIN {right}')

        assert this is not None and that is not None

        if constraint.fields == constraint.foreign_fields:
            self.add(['USING', '('])

            for field in constraint.fields:
                self.add(field)
                self.add(',')

            self._stat.pop()
            self.add(')')

        else:
            self.add(['ON', '('])
            for af, bf in zip(constraint.fields, constraint.foreign_fields):
                self.add([f'{this}.{af} = {that}.{bf}'])
                self.add('AND')

            self._stat.pop()
            self.add(')')

        self._involved.append(right)

    @catch_error
    def group_by(self, *by) -> Self:
        """
        ``GROUP BY`` https://www.sqlite.org/lang_select.html#resultset
        """
        if len(by) == 0:
            raise RuntimeError()

        self.add('GROUP BY')
        for field in by:
            if isinstance(field, (int, float, bool)):
                self.add(str(field))
            elif isinstance(field, str):
                self.add(repr(field))
            elif isinstance(field, Field):
                self.add(f'{field.table_name}.{field.name}')
            elif isinstance(field, SqlField):
                self.add(f'{field.table_name}.{field.name}')
            elif isinstance(field, SqlAlias):
                self.add(field._name)
            elif isinstance(field, SqlExpr):
                self.add(field)
            else:
                raise TypeError('GROUP BY ' + repr(field))
            self.add(',')

        self._stat.pop()
        return self

    @catch_error
    def having(self, *exprs: Union[bool, SqlExpr]) -> Self:
        """
        ``HAVING`` https://www.sqlite.org/lang_select.html#resultset
        """
        if len(exprs) == 0:
            return self

        from .func_stat import and_
        self.add('HAVING')
        self.add(and_(*exprs))

        return self

    @catch_error
    def intersect(self, stat: SqlStat) -> Self:
        """
        ``INTERSECT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('INTERSECT')
        self._stat.extend(stat._stat)
        stat._connection = None
        stat._stat = None
        return self

    def __and__(self, other: SqlStat) -> Self:
        """
        ``INTERSECT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.intersect(other)

    @catch_error
    def union(self, stat: SqlStat, all=False) -> Self:
        """
        ``UNION`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('UNION')
        if all:
            self.add('ALL')
        self._stat.extend(stat._stat)
        stat._connection = None
        stat._stat = None
        return self

    def __or__(self, other: SqlStat) -> Self:
        """
        ``UNION`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.union(other)

    @catch_error
    def except_(self, stat: SqlStat) -> Self:
        """
        ``EXCEPT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('EXCEPT')
        self._stat.extend(stat._stat)
        stat._connection = None
        stat._stat = None
        return self

    def __sub__(self, other: SqlStat) -> Self:
        """
        ``EXCEPT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.except_(other)


class SqlReturnStat:
    @catch_error
    def returning(self, *expr: Union[str, Any]) -> Self:
        zelf = cast(SqlStat, self)
        zelf.add('RETURNING')
        if len(expr) == 0:
            zelf.add('*')
        else:
            fields = []
            for exp in expr:
                if isinstance(exp, str):
                    zelf.add(exp)
                    fields.append(exp)
                elif isinstance(exp, SqlField):
                    zelf.add(exp.name)
                    fields.append(exp.name)
                elif isinstance(exp, SqlAlias) and isinstance(field := exp._value, SqlField):
                    zelf.add([field.name, 'AS', exp._name])
                    fields.append(exp._name)
                elif isinstance(exp, SqlAlias) and isinstance(_expr := exp._value, SqlExpr):
                    zelf.add(SqlRemovePlaceHolder(_expr))

                    zelf.add(['AS', exp._name])
                    fields.append(exp._name)
                else:
                    raise TypeError(f'RETURNING ({exp})')

                zelf.add(',')
            zelf._stat.pop()

            if isinstance(self, SqlInsertStat):
                self._return_table = None
                self._fields = fields
            else:
                self.table = None

        return self


class SqlInsertStat(SqlStat[T], SqlReturnStat, Generic[T]):
    def __init__(self, table: type[T], fields: list[str] = None, *, named: bool = False):
        super().__init__(table)
        self._fields = fields
        self._used_fields: list[str] | None = None

        # when
        #   None: `VALUES` set
        #   'DEFAULT': `DEFAULT VALUES` unset
        #   {field->value}: `DEFAULT VALUES` set
        self._values: Union[Literal['DEFAULT'], dict[str, Any], None] = {
            name: SqlLiteral('?')
            for name in table_field_names(table)
        }
        self._named = named

        self._returning = False
        self._return_table = table

    @overload
    def select_from(self, table: type[T], *, distinct: bool = False) -> SqlSelectStat[T]:
        pass

    @overload
    def select_from(self, *field, distinct: bool = False,
                    from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat[tuple]:
        pass

    @catch_error
    def select_from(self, *args, distinct: bool = False,
                    from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat[T]:
        if isinstance(self._values, str):
            raise RuntimeError()

        self._connection = None

        self.add('(')
        for i, field in enumerate(self._values):
            if i > 0:
                self.add(',')
            self.add(field)
        self.add(')')

        from .stat_start import select_from
        ret = select_from(*args, distinct=distinct, from_table=from_table)
        ret._stat = [*self._stat, *ret._stat]

        return ret

    @catch_error
    def values(self, *args, **kwargs: Union[str, SqlExpr]) -> Self:
        # TODO support direct set values
        if self._values is None or isinstance(self._values, str):
            raise RuntimeError()

        for expr in args:
            if isinstance(expr, SqlCompareOper) and expr.oper == '=' and isinstance(expr.left, SqlField):
                self._values[expr.left.field.name] = expr.right
            else:
                raise TypeError(expr)

        for field, expr in kwargs.items():
            if field not in self._values:
                raise RuntimeError(f'{table_name(self.table)}.{field} not found')
            self._values[field] = expr
        return self

    def defaults(self) -> Self:
        """insert default values"""
        self._values = 'DEFAULT'
        return self

    @catch_error
    def _set_values(self):
        if isinstance(self._values, str):
            self.add(['DEFAULT', 'VALUES'])
        elif isinstance(self._values, dict):
            self._use_fields = []

            if self._fields is None:
                fields = list(self._values.keys())
            else:
                fields = self._fields

            self.add(['VALUES', '('])

            for field in fields:
                value = self._values[field]
                if isinstance(value, (int, float, bool, str)):
                    self.add(repr(value))
                elif isinstance(value, SqlLiteral):
                    if value.value == '?':
                        if self._named:
                            self.add(f':{field}')
                        else:
                            self.add('?')
                        self._use_fields.append(field)
                    else:
                        self.add(value.value)
                elif isinstance(value, SqlPlaceHolder):
                    self.add(repr(value.value))
                elif isinstance(value, SqlStat):
                    self.add(value)
                elif isinstance(value, SqlExpr):
                    self.add(SqlRemovePlaceHolder(value))
                else:
                    raise TypeError(repr(value))
                self.add(',')
            self._stat.pop()
            self.add(')')

        self._values = None

    @catch_error
    def on_conflict(self, *conflict, where: Union[bool, SqlCompareOper] = None) -> SqlUpsertStat[T]:
        self._set_values()
        return SqlUpsertStat(self, *conflict, where=where)

    @catch_error
    def returning(self, *expr: Union[str, SqlExpr]) -> SqlInsertStat[tuple]:
        self._set_values()
        super().returning(*expr)
        self._returning = True
        return self

    def build(self) -> tuple[str, list]:
        self._set_values()
        return super().build()

    def submit(self, parameter: list[T] = ()) -> Cursor[T]:
        if (connection := self._connection) is None:
            raise RuntimeError('Do not in a connection context')

        self._set_values()

        from .table import table_class
        try:
            table_cls = table_class(self.table)
        except AttributeError:
            pass
        else:
            if self._named:
                def mapper(p):
                    if isinstance(p, self.table):
                        return {f: v for f, v in zip(self._use_fields, table_cls.table_seq(p, self._fields))}
                    return dict(p)
            else:
                def mapper(p):
                    if isinstance(p, self.table):
                        return table_cls.table_seq(p, self._use_fields)
                    return tuple(p)

            parameter = list(map(mapper, parameter))

        if len(parameter):
            if self._returning:
                if len(parameter) > 1:
                    raise RuntimeError('only support return one data')
                cur = connection.execute(self, parameter[0])
            else:
                cur = connection.execute_batch(self, parameter)
        else:
            cur = connection.execute(self, parameter)

        ret = Cursor(connection, cur, self._return_table)
        self._connection = None
        return ret


class SqlUpsertStat(Generic[T]):
    """

    https://www.sqlite.org/syntax/upsert-clause.html

    """

    def __init__(self, stat: SqlInsertStat[T], *conflict, where: Union[bool, SqlCompareOper] = None):
        self._stat = stat
        self._conflict = conflict
        self._stat.add(['ON', 'CONFLICT'])
        self._do_where = False

        if len(conflict):
            from .stat_start import select_from_fields
            table, fields = select_from_fields(*conflict)
            if table is None:
                table = stat.table
            elif table != stat.table:
                raise RuntimeError()

            self._stat.add('(')
            for i, field in enumerate(fields):
                if i > 0:
                    self._stat.add(',')

                if isinstance(field, SqlField):
                    if field.table != table:
                        raise RuntimeError(f'field {field.table_name}.{field.name} not belong to {table.__name__}')
                    self._stat.add(f'{field.name}')
                elif isinstance(field, SqlLiteral) and isinstance(field.value, str):
                    self._stat.add(field.value)
                else:
                    raise RuntimeError(f'ON CONFLICT ({field}:{type(field).__name__})')

            if where is not None:
                self._stat.add('WHERE')
                self._stat.add(SqlRemovePlaceHolder(where))
            self._stat.add(')')

    @catch_error(attr='_stat')
    def do_nothing(self) -> SqlInsertStat[T]:
        self._stat.add(['DO', 'NOTHING'])
        return self._stat

    @catch_error(attr='_stat')
    def do_update(self, *args: Union[bool, SqlCompareOper], where: Union[bool, SqlCompareOper] = None) -> SqlInsertStat[T]:
        self._stat.add(['DO', 'UPDATE', 'SET'])

        self._stat.add(SqlRemovePlaceHolder(SqlVarArgOper(',', [
            SqlCompareOper.as_set_expr(it) for it in args
        ])))

        if where is not None:
            self._stat.add('WHERE')
            self._stat.add(SqlRemovePlaceHolder(where))
        return self._stat


class SqlUpdateStat(SqlStat[T], SqlWhereStat, SqlLimitStat, SqlReturnStat, Generic[T]):
    @catch_error
    def from_(self, query: Union[SqlStat, SqlAlias[SqlSubQuery]]) -> Self:
        self.add('FROM')
        if isinstance(query, SqlStat):
            self.add(query)
        elif isinstance(query, SqlSubQuery):
            self.add(query.stat)
        elif isinstance(query, SqlAlias) and isinstance(query._value, SqlSubQuery):
            self.add(query._value.stat)
            self.add(['AS', query._name])
        else:
            raise TypeError(repr(query))
        return self


class SqlDeleteStat(SqlStat[T], SqlWhereStat, SqlLimitStat, SqlReturnStat, Generic[T]):
    pass



import numpy as np

__all__ = ['interp_timestamp',
           'interp1d_nan']


def interp_timestamp(timestamp: np.ndarray,
                     t0: float,
                     t1: float,
                     sampling_rate: float) -> tuple[np.ndarray, np.ndarray]:
    """
    Create interpolated time and value(01) array from event timestamp

    `Dimension parameters`:

        T = number of timestamp

        N' = number of interpolated sample points

    :param timestamp: Array of event timestamps in seconds. `Array[float, T]`
    :param t0: Start time in seconds
    :param t1: End time in seconds
    :param sampling_rate: Sampling rate in Hz
    :return: Tuple of time array (`Array[float, N']`) and event array (`Array[float, N']`)
    """

    n = int((t1 - t0) * sampling_rate)
    tt = np.linspace(t0, t1, num=n)

    sig = np.zeros(n, dtype=np.int8)
    indices = ((timestamp - t0) * sampling_rate).astype(int)
    indices = indices[(indices >= 0) & (indices < n)]
    sig[indices] = 1

    return tt, sig


def interp1d_nan(arr: np.ndarray) -> np.ndarray:
    """
    Interpolates missing values (NaNs) in a 1-dimensional NumPy array.

    :param arr: Input array with potential NaNs. `Array[float, N]`
    :return: Array with NaNs interpolated. `Array[float, N]`
    """
    n = len(arr)
    x = np.arange(n)
    isnan = np.isnan(arr)

    if isnan.any():
        if isnan.all():
            raise ValueError('Input array contains only NaNs')

        arr_nonan = arr[~isnan]
        x_nonan = x[~isnan]
        arr[isnan] = np.interp(x[isnan], x_nonan, arr_nonan)

    arr_clean = arr[~np.isnan(arr)]

    return arr_clean



"""
SQL functions.

covers
* https://www.sqlite.org/lang_corefunc.html
* https://www.sqlite.org/lang_aggfunc.html
* https://www.sqlite.org/lang_mathfunc.html
"""
from typing import overload, TypeVar, Any

from typing_extensions import LiteralString

from . import expr
from .func_dec import as_func_expr

__all__ = [
    'abs', 'changes', 'char', 'coalesce', 'format',
    'hex', 'ifnull', 'iif', 'instr', 'last_insert_rowid', 'length', 'likelihood', 'likely', 'lower', 'ltrim',
    'max', 'min', 'nullif', 'octet_length', 'printf', 'quote', 'random', 'randomblob', 'replace', 'round', 'rtrim',
    'sign', 'sqlite_source_id', 'sqlite_version', 'substr', 'total_changes', 'trim', 'typeof',
    'unhex', 'unicode', 'unlikely', 'upper', 'zeroblob',
    'avg', 'count', 'group_concat', 'sum',
    'acos', 'acosh', 'asin',
    'asinh', 'atan', 'atan2', 'atanh', 'ceil', 'cos', 'cosh', 'degrees', 'exp', 'floor', 'ln', 'log', 'log2', 'log10',
    'mod', 'pi', 'pow', 'radians', 'sin', 'sinh', 'sqrt', 'tan', 'tanh', 'trunc',
]

T = TypeVar('T')
N = TypeVar('N', int, float)


# noinspection PyShadowingBuiltins
@as_func_expr
def abs(x: N) -> N:
    """https://www.sqlite.org/lang_corefunc.html#abs"""
    import builtins
    return builtins.abs(x)


@as_func_expr
def changes() -> int:
    """https://www.sqlite.org/lang_corefunc.html#changes"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr
def char(*x) -> str:
    """https://www.sqlite.org/lang_corefunc.html#char"""
    raise NotImplementedError()


@as_func_expr
def coalesce(*x: T) -> T | None:
    """https://www.sqlite.org/lang_corefunc.html#coalesce"""
    for y in x:
        if y is not None:
            return y
    return None


# https://www.sqlite.org/lang_corefunc.html#concat by func_stat
# https://www.sqlite.org/lang_corefunc.html#glob by func_stat

# noinspection PyShadowingBuiltins
@as_func_expr
def format(fmt: LiteralString, *x) -> str:
    """https://www.sqlite.org/lang_corefunc.html#format"""
    return fmt % x


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def hex(x: bytes) -> str:
    """https://www.sqlite.org/lang_corefunc.html#hex"""
    raise NotImplementedError()


@as_func_expr
def ifnull(x: T, y: T) -> T:
    """
    ``x?:y``

    https://www.sqlite.org/lang_corefunc.html#ifnull

    """
    if x is not None:
        return x
    return y


@as_func_expr
def iif(x, y: T, z: T) -> T:
    """
    ``x?y:z``

    https://www.sqlite.org/lang_corefunc.html#iif
    """
    if x:
        return y
    else:
        return z


@as_func_expr
def instr(x: str, y: str) -> int:
    """
    ``x.find(y)``

    https://www.sqlite.org/lang_corefunc.html#instr
    """
    return x.find(y) + 1


@as_func_expr
def last_insert_rowid() -> int:
    """https://www.sqlite.org/lang_corefunc.html#last_insert_rowid"""
    raise NotImplementedError()


@as_func_expr
def length(x: str | bytes) -> int:
    """https://www.sqlite.org/lang_corefunc.html#length"""
    return len(x)


# https://www.sqlite.org/lang_corefunc.html#like by func_stat

@as_func_expr
def likelihood(x: T, y: float) -> T:
    """https://www.sqlite.org/lang_corefunc.html#likelihood"""
    return x


@as_func_expr
def likely(x: T) -> T:
    """https://www.sqlite.org/lang_corefunc.html#likely"""
    return x


# TODO https://www.sqlite.org/lang_corefunc.html#load_extension

@as_func_expr
def lower(x: str) -> str:
    """https://www.sqlite.org/lang_corefunc.html#lower"""
    return x.lower()


@as_func_expr
def ltrim(x: str, y: str = None) -> str:
    """https://www.sqlite.org/lang_corefunc.html#ltrim"""
    return x.lstrip(y)


# noinspection PyShadowingBuiltins
@overload
def max(x: N) -> N:
    raise NotImplementedError()


# noinspection PyShadowingBuiltins
@overload
def max(x: N, *other: N) -> N:
    raise NotImplementedError()


# noinspection PyShadowingBuiltins
def max(x, *other):
    """
    https://www.sqlite.org/lang_corefunc.html#max
    https://www.sqlite.org/lang_aggfunc.html#max_agg
    """
    if len(other) == 0:
        return expr.SqlAggregateFunc('MAX', x)
    else:
        return expr.SqlFuncOper('MAX', _max, x, *other)


def _max(*other):
    import builtins
    return builtins.max(*other)


# noinspection PyShadowingBuiltins
@overload
def min(x: N) -> N:
    raise NotImplementedError()


# noinspection PyShadowingBuiltins
@overload
def min(x: N, *other: N) -> N:
    raise NotImplementedError()


# noinspection PyShadowingBuiltins
def min(x, *other):
    """
    https://www.sqlite.org/lang_corefunc.html#min
    https://www.sqlite.org/lang_aggfunc.html#min_agg
    """
    if len(other) == 0:
        return expr.SqlAggregateFunc('MIN', x)
    else:
        return expr.SqlFuncOper('MIN', _min, x, *other)


def _min(*other):
    import builtins
    return builtins.min(*other)


@as_func_expr
def nullif(x: T, y: T) -> T | None:
    """
    ``x if x != y else None``

    https://www.sqlite.org/lang_corefunc.html#nullif
    """
    return x if x != y else None


# noinspection PyUnusedLocal
@as_func_expr
def octet_length(x: str) -> int:
    """https://www.sqlite.org/lang_corefunc.html#octet_length"""
    raise NotImplementedError()


@as_func_expr
def printf(fmt: LiteralString, *x) -> str:
    """https://www.sqlite.org/lang_corefunc.html#printf"""
    return fmt % x


# noinspection PyUnusedLocal
@as_func_expr
def quote(x: str) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#quote"""
    raise NotImplementedError()


@as_func_expr
def random() -> int:
    """https://www.sqlite.org/lang_corefunc.html#random"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr
def randomblob(n: int) -> bytes:
    """https://www.sqlite.org/lang_corefunc.html#randomblob"""
    raise NotImplementedError()


@as_func_expr
def replace(x: str, y: str, z: str) -> str:
    """https://www.sqlite.org/lang_corefunc.html#replace"""
    if len(y) == 0:
        return x
    else:
        return x.replace(y, z)


# noinspection PyShadowingBuiltins
@as_func_expr
def round(x: float, y: int = None) -> float:
    """https://www.sqlite.org/lang_corefunc.html#round"""
    import builtins
    if y is None:
        y = 0
    return builtins.round(x, y)


@as_func_expr
def rtrim(x: str, y: str = None) -> str:
    """https://www.sqlite.org/lang_corefunc.html#rtrim"""
    return x.rstrip(y)


@overload
def sign(x: N) -> int:
    pass


@overload
def sign(x: Any) -> None:
    pass


@as_func_expr
def sign(x):
    """https://www.sqlite.org/lang_corefunc.html#sign"""
    if isinstance(x, (int, float)):
        if x > 0:
            return 1
        elif x < 0:
            return -1
        else:
            return 0
    return None


# TODO https://www.sqlite.org/lang_corefunc.html#soundex
# https://www.sqlite.org/lang_corefunc.html#sqlite_compileoption_get by Connection.sqlite_compileoption_get
# https://www.sqlite.org/lang_corefunc.html#sqlite_compileoption_used by Connection.sqlite_compileoption_used
# TODO https://www.sqlite.org/lang_corefunc.html#sqlite_offset

@as_func_expr
def sqlite_source_id() -> str:
    """https://www.sqlite.org/lang_corefunc.html#sqlite_source_id"""
    raise NotImplementedError()


@as_func_expr
def sqlite_version() -> str:
    """https://www.sqlite.org/lang_corefunc.html#sqlite_version"""
    raise NotImplementedError()


@as_func_expr
def substr(x: str, y: int, z: int = None) -> str:
    """https://www.sqlite.org/lang_corefunc.html#substr"""
    if z is None:
        if y > 0:
            return x[y - 1:]
        else:
            return x[y:]
    elif z > 0:
        if y > 0:
            return x[y - 1:y + z - 1]
        else:
            return x[y:y + z]
    else:
        if y > 0:
            return x[y + z - 1:y - 1]
        else:
            return x[y + z:y]


@as_func_expr
def total_changes() -> int:
    """https://www.sqlite.org/lang_corefunc.html#total_changes"""
    raise NotImplementedError()


@as_func_expr
def trim(x: str, y: str = None) -> str:
    """https://www.sqlite.org/lang_corefunc.html#trim"""
    return x.strip(y)


# noinspection PyUnusedLocal
@as_func_expr
def typeof(x) -> str:
    """https://www.sqlite.org/lang_corefunc.html#typeof"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr
def unhex(x: str, y=None) -> bytes:
    """https://www.sqlite.org/lang_corefunc.html#unhex"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr
def unicode(x: str) -> int:
    """https://www.sqlite.org/lang_corefunc.html#unicode"""
    raise NotImplementedError()


@as_func_expr
def unlikely(x: T) -> T:
    """https://www.sqlite.org/lang_corefunc.html#unlikely"""
    return x


@as_func_expr
def upper(x: str) -> str:
    """https://www.sqlite.org/lang_corefunc.html#upper"""
    return x.upper()


# noinspection PyUnusedLocal
@as_func_expr
def zeroblob(n: int) -> bytes:
    """https://www.sqlite.org/lang_corefunc.html#zeroblob"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def avg(x) -> float:
    """https://www.sqlite.org/lang_aggfunc.html#avg"""
    raise NotImplementedError()


def count(x=None) -> int:
    """https://www.sqlite.org/lang_aggfunc.html#count"""
    if x is None:
        return expr.SqlAggregateFunc('COUNT', expr.SqlLiteral('*'))
    else:
        return expr.SqlAggregateFunc('COUNT', x)


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def group_concat(x, y=None) -> str:
    """https://www.sqlite.org/lang_aggfunc.html#group_concat"""
    raise NotImplementedError()


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def sum(x) -> float:
    """https://www.sqlite.org/lang_aggfunc.html#sumunc"""
    raise NotImplementedError()


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def total(x) -> int:
    """https://www.sqlite.org/lang_aggfunc.html#sumunc"""
    raise NotImplementedError()


@as_func_expr
def acos(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#acos"""
    import math
    return math.acos(x)


@as_func_expr
def acosh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#acosh"""
    import math
    return math.acosh(x)


@as_func_expr
def asin(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#asin"""
    import math
    return math.asin(x)


@as_func_expr
def asinh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#asinh"""
    import math
    return math.asinh(x)


@as_func_expr
def atan(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#atan"""
    import math
    return math.atan(x)


@as_func_expr
def atan2(y: float, x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#atan2"""
    import math
    return math.atan2(y, x)


@as_func_expr
def atanh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#atanh"""
    import math
    return math.atanh(x)


@as_func_expr
def ceil(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#ceil"""
    import math
    return math.ceil(x)


@as_func_expr
def cos(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#cos"""
    import math
    return math.cos(x)


@as_func_expr
def cosh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#cosh"""
    import math
    return math.cosh(x)


@as_func_expr
def degrees(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#degrees"""
    import math
    return math.degrees(x)


@as_func_expr
def exp(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#exp"""
    import math
    return math.exp(x)


@as_func_expr
def floor(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#floor"""
    import math
    return math.floor(x)


@as_func_expr
def ln(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#ln"""
    import math
    return math.log(x)


@as_func_expr
def log2(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#log2"""
    import math
    return math.log2(x)


@as_func_expr
def log10(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#log"""
    import math
    return math.log10(x)


@as_func_expr
def log(b: float, x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#log"""
    import math
    return math.log(x, b)


@as_func_expr
def mod(x: float, y: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#mod"""
    return x % y


@as_func_expr
def pi() -> float:
    """https://www.sqlite.org/lang_mathfunc.html#pi"""
    import math
    return math.pi


@as_func_expr
def pow(x: float, y: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#pow"""
    import math
    return math.pow(x, y)


@as_func_expr
def radians(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#radians"""
    import math
    return math.radians(x)


@as_func_expr
def sin(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#sin"""
    import math
    return math.sin(x)


@as_func_expr
def sinh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#sinh"""
    import math
    return math.sinh(x)


@as_func_expr
def sqrt(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#sqrt"""
    import math
    return math.sqrt(x)


@as_func_expr
def tan(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#tan"""
    import math
    return math.tan(x)


@as_func_expr
def tanh(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#tanh"""
    import math
    return math.tanh(x)


@as_func_expr
def trunc(x: float) -> float:
    """https://www.sqlite.org/lang_mathfunc.html#trunc"""
    import math
    return math.trunc(x)



import pickle
import shutil
from contextlib import contextmanager
from pathlib import Path
from typing import Any, ContextManager, TYPE_CHECKING

import gdown
import numpy as np
import polars as pl
from PIL import Image

from neuralib.io import NEUROLIB_DATASET_DIRECTORY
from neuralib.typing import PathLike

if TYPE_CHECKING:
    from neuralib.tracking.deeplabcut import DeepLabCutDataFrame
    from neuralib.imaging.suite2p import Suite2PResult
    from neuralib.imaging.scanbox import SBXInfo
    from neuralib.tracking.facemap import FaceMapResult
    from neuralib.model.rastermap import RasterMapResult
    from neuralib.scan.czi import CziScanner
    from neuralib.scan.lsm import TiffScanner

__all__ = [
    'google_drive_file',
    'google_drive_folder',
    'clean_all_cache_dataset',
    #
    'load_example_rois',
    'load_example_rois_image',
    'load_example_rois_dir',
    'load_example_dorsal_cortex',
    #
    'load_ephys_meta',
    'load_ephys_data',
    'load_npx2_reconstruction',
    #
    'load_example_scanbox',
    'load_example_suite2p_result',
    'load_example_rastermap_2p_result',
    'load_example_rastermap_2p_cache',
    'load_example_retinotopic_data',
    #
    'load_example_dlc_h5',
    'load_example_dlc_csv',
    'load_example_facemap_pupil',
    'load_example_facemap_keypoints',
    #
    'load_example_lsm',
    'load_example_czi'
]


@contextmanager
def google_drive_file(file_id: str,
                      *,
                      quiet: bool = False,
                      output_dir: PathLike | None = None,
                      rename_file: str | None = None,
                      cached: bool = False,
                      invalid_cache: bool = False) -> ContextManager[Path]:
    """
    Download file from Google Drive. If not ``cached``, then delete afterward.

    :param file_id: Google Drive file ID used to identify the file to be downloaded.
    :param quiet: Boolean flag to suppress output from the download process.
    :param output_dir: Directory path where the downloaded file will be saved.
    :param rename_file: Optional string to rename the downloaded file.
    :param cached: Boolean flag to retain the downloaded file after usage.
    :param invalid_cache: Boolean flag to force re-download even if the file exists in the cache.
    :return: A context manager yielding the path to the downloaded file.
    """
    if rename_file is not None:
        file = rename_file
    else:
        file = file_id

    if output_dir is None:
        output_dir = NEUROLIB_DATASET_DIRECTORY

    output_dir.mkdir(exist_ok=True, parents=True)
    output_file = output_dir / file

    try:
        if output_file.exists() and not invalid_cache:
            yield output_file
        else:
            url = f"https://drive.google.com/uc?id={file_id}"
            gdown.download(url, str(output_file), quiet=quiet)
            yield output_file
    finally:
        if not cached:
            output_file.unlink(missing_ok=True)


@contextmanager
def google_drive_folder(folder_id: str,
                        *,
                        quiet: bool = False,
                        output_dir: Path | None = None,
                        rename_folder: str | None = None,
                        cached: bool = False,
                        invalid_cache: bool = False) -> ContextManager[Path]:
    """
    Download a entire folder from Google Drive. If not ``cached``, then delete afterward.

    :param folder_id: Google Drive folder ID to download.
    :param quiet: If True, suppress the gdown output.
    :param output_dir: Directory where the folder should be downloaded. If None, a temporary directory is used.
    :param rename_folder: Optional new name for the downloaded folder.
    :param cached: If True, keep the downloaded folder for future use.
    :param invalid_cache: If True, force re-download even if cached data exists.
    :return: A context manager yielding the path to the downloaded folder.
    """
    if rename_folder is not None:
        folder_name = rename_folder
    else:
        folder_name = folder_id

    if output_dir is None:
        output_dir = NEUROLIB_DATASET_DIRECTORY / folder_name

    try:
        if output_dir.exists() and any(output_dir.iterdir()) and not invalid_cache:
            yield output_dir
        else:
            output_dir.mkdir(exist_ok=True, parents=True)
            gdown.download_folder(id=folder_id, output=str(output_dir), quiet=quiet)
            yield output_dir
    finally:
        if not cached:
            shutil.rmtree(output_dir, ignore_errors=True)


def clean_all_cache_dataset():
    """removes the cache dataset directory"""
    shutil.rmtree(NEUROLIB_DATASET_DIRECTORY)
    print(f'[REMOVE]: {NEUROLIB_DATASET_DIRECTORY}')


# ========== #
# Atlas Data #
# ========== #

def load_example_rois(**kwargs) -> pl.DataFrame:
    """
    Load example ROIs dataframe

    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: A Polars DataFrame containing the example ROIs data
    """
    with google_drive_file('1dKpZt6eF4szvl7svWRdBQkOfTVLQi4Xg', **kwargs) as file:
        return pl.read_csv(file)


def load_example_rois_image(**kwargs) -> np.ndarray:
    """
    Load example ROIs image array

    :param kwargs: Additional arguments to be passed to the `google_drive_file` context manager.
    :return: An example imaging array with labeled ROIs
    """
    with google_drive_file('1-ZFC7Fd6IgwbY6X8oetvpGjZaBBiwg0t', **kwargs) as file:
        return np.array(Image.open(file))


def load_example_rois_dir(**kwargs) -> Path:
    """Load a directory containing multiple example ROIs image array"""
    with google_drive_folder('1tj36-lzOArjwMsZlhbyy_B19ffl50J98', **kwargs) as d:
        return d


def load_example_dorsal_cortex(color: bool = False, **kwargs) -> np.ndarray:
    """Load dorsal cortex image array from the source svg

    .. seealso::

        :meth:`~neuralib.atlas.data.get_dorsal_cortex`

    :param color:

    """
    if color:
        file_id = '1Cujx3GGFZxEq0-isRlA_Ac-Puy6ml8IA'
    else:
        file_id = '1OEPpIIl8SszDJXO_1ADka-5pB556WNxc'

    with google_drive_file(file_id, **kwargs) as file:
        return np.array(Image.open(file))


# ========== #
# Ephys Data #
# ========== #

def load_ephys_meta(**kwargs):
    with google_drive_file('1gB2MegBAJbobEudx2pZmVe0LBQiU8HFU', **kwargs) as _:
        pass


def load_ephys_data(**kwargs):
    with google_drive_file('1U0xAchQagyXRT72M68fRQ4JsRQeW9q5d', **kwargs) as _:
        pass


def load_npx2_reconstruction(**kwargs) -> pl.DataFrame:
    """
    Example of NeuroPixel2 reconstruction data

    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: A Polars DataFrame containing the example DiI labelled traces ROIs
    """
    with google_drive_file('1fRvMNHhGgh5KP3CgGm6CMFth1qIAmwfh', **kwargs) as file:
        return pl.read_csv(file)


# ============ #
# Imaging Data #
# ============ #

def load_example_scanbox(**kwargs) -> 'SBXInfo':
    """
    Load example ScanBox data

    :param kwargs: Additional keyword arguments pass to ``google_drive_file`` to customize the loading behavior.
    :return: An instance of ``SBXInfo`` loaded from the specified Google Drive file.
    """
    from neuralib.imaging.scanbox import SBXInfo
    with google_drive_file('1Gcz_xRVWQJ9QMxq3vzZS8VruSbNiuh_s', **kwargs) as file:
        return SBXInfo.load(file)


def load_example_suite2p_result(**kwargs) -> 'Suite2PResult':
    """
    Load example Suite2P data

    :param kwargs: Additional keyword arguments pass to ``google_drive_folder`` to customize the loading behavior.
    :return: An instance of ``Suite2PResult`` loaded with data from the specified Google Drive folder.
    """
    from neuralib.imaging.suite2p import read_suite2p
    with google_drive_folder('1iVImr_rIywWhCiBDYhcphcSODaWJrhy7', **kwargs) as suite2p_dir:
        return read_suite2p(suite2p_dir)


def load_example_rastermap_2p_result(**kwargs) -> 'RasterMapResult':
    """Load example rastermap 2P data"""
    from neuralib.model.rastermap import read_rastermap
    with google_drive_file('1KSic4sXyF3hTgQbGijMpa3D3TGJUU097', **kwargs) as file:
        return read_rastermap(file)


def load_example_rastermap_2p_cache(**kwargs) -> dict[str, Any]:
    """
    Load example rastermap 2P cache pickle file
    :param kwargs: Additional arguments to be passed to the `google_drive_file` context manager.
    :return: A dictionary containing the 2-photon rastermap data cache
    """
    with google_drive_file('1SuzUhkbcdBY71dCnxjnT4pmQOsiOULBC', **kwargs) as file:
        with file.open('rb') as f:
            return pickle.load(f)


def load_example_retinotopic_data(**kwargs) -> Path:
    """Load example retinotopic tiff file path"""
    with google_drive_file('1J8iqP_EBaknNJehRUw3nwp7lEw2_UFXz',
                           quiet=True,
                           cached=True,
                           rename_file='retinotopic.tiff', **kwargs) as file:
        return Path(file)


# ========== #
# Behavioral #
# ========== #

def load_example_dlc_h5(**kwargs) -> 'DeepLabCutDataFrame':
    """Load example Deeplabcut h5"""
    with google_drive_file('1JNhx6Dpe8beP8vnh0yF3o3vY2DfUM-8A', rename_file='test_dlc.h5', **kwargs) as h5:
        from neuralib.tracking.deeplabcut.core import read_dlc
        with google_drive_file('1juICYcrXa7Vk-fQSBBSg2QcP9DGyHO2E', rename_file='test_dlc.pickle', **kwargs) as meta:
            return read_dlc(h5, meta)


def load_example_dlc_csv(**kwargs) -> 'DeepLabCutDataFrame':
    """Load example Deeplabcut csv"""
    from neuralib.tracking.deeplabcut.core import read_dlc
    with google_drive_file('1R2Ze5xjWlavcKvu45JOH3_QkOD4SSkVN', rename_file='test_dlc.csv', **kwargs) as csv:
        with google_drive_file('1juICYcrXa7Vk-fQSBBSg2QcP9DGyHO2E', rename_file='test_dlc.pickle', **kwargs) as meta:
            return read_dlc(csv, meta)


def load_example_facemap_pupil(**kwargs) -> 'FaceMapResult':
    """Load example facemap pupil data"""
    from neuralib.tracking.facemap import read_facemap
    with google_drive_folder('1cacH5DWLmYqh_7PLwqEasmER_TfKgZ1b', **kwargs) as pupil_dir:
        return read_facemap(pupil_dir)


def load_example_facemap_keypoints(**kwargs) -> 'FaceMapResult':
    """Load example facemap keypoint data"""
    from neuralib.tracking.facemap import read_facemap
    with google_drive_folder('1FWz70HE_hQuhE6K9hoO_y1OgeG11NsGM', **kwargs) as pupil_dir:
        return read_facemap(pupil_dir)


# ============= #
# Confocal Scan #
# ============= #

def load_example_lsm(**kwargs) -> 'TiffScanner':
    """load example lsm file"""
    from neuralib.scan.lsm import lsm_file
    with google_drive_file('1beq6PCY8XmZjyWiOk2-KkcXLrmbLpWmS', rename_file='test.lsm', **kwargs) as file:
        with lsm_file(file) as lsm:
            return lsm


def load_example_czi(**kwargs) -> 'CziScanner':
    """load example czi file"""
    from neuralib.scan.czi import czi_file
    with google_drive_file('1gSPz_a7kCZ3UQABC-v-_sNnAxhfY_ly0', rename_file='test.czi', **kwargs) as file:
        with czi_file(file) as czi:
            return czi



from pathlib import Path
from typing import Any

import numpy as np
import sbxreader
from rich.pretty import pprint

from argclz import AbstractParser, argument, pos_argument, validator, int_tuple_type
from neuralib.imaging.scanbox import SBXInfo
from neuralib.typing import PathLike

__all__ = ['ScanBoxView',
           'ScanBoxViewOptions']

from neuralib.util.utils import uglob


class ScanBoxView:
    """Scanbox info and data

    `Dimension parameters`:

        F = number of frames

        P = number of optical planes

        C = number of PMT channels

        W = FOV width

        H = FOV height
    """

    info: SBXInfo
    """:class:`~neuralib.imaging.scanbox.core.SBXInfo`"""

    sbx_map: sbxreader.sbx_memmap
    """`Array[float, [F, P, C, W, H]]`"""

    def __init__(self, directory: PathLike):
        """
        :param directory: directory contain ``.sbx`` & ``.mat`` files
        """
        if not Path(directory).is_dir():
            raise NotADirectoryError(f'{directory}')

        sbx = uglob(directory, '*.sbx')
        self.sbx_map = sbxreader.sbx_memmap(sbx)

        info = uglob(directory, '*.mat')
        self.info = SBXInfo.load(info)

    @property
    def meta(self) -> dict[str, Any]:
        """scanbox meta information"""
        return self.info.asdict()

    @property
    def version(self) -> int:
        """scanbox version"""
        return int(self.info.scanbox_version)

    @property
    def height(self) -> int:
        """fov height"""
        return self.info.sz[0]

    @property
    def width(self) -> int:
        """fov width"""
        return self.info.sz[1]

    @property
    def n_planes(self) -> int:
        """number of optical imaging planes"""
        if not len(self.info.otwave) and self.info.volscan == 0:
            return 1
        elif self.info.volscan == 1:
            return len(self.info.otwave)
        raise RuntimeError('')

    @property
    def n_channels(self) -> int:
        """sbx issue, a bit of hard-coded"""
        if self.info.nchan is None:
            if self.info.channels == 2:
                return 1
            elif self.info.channels == 1:
                return 2
            else:
                raise RuntimeError('')
        else:
            return self.info.nchan

    @property
    def n_frames(self) -> int:
        """number of frames per plane"""
        return int(self.info.config.frames / self.n_planes)

    def show(self,
             frames: slice | np.ndarray | None,
             plane: int,
             channel: int):
        """play the selected frames using customized CV2 player

        :param frames: selected frames. If None, play all sequences
        :param plane: number of optical planes
        :param channel: number of PMT channel
        """
        from neuralib.imglib.labeller import SequenceLabeller

        if frames is None:
            frames = np.arange(0, self.n_frames)

        data = self.sbx_map[frames, plane, channel, :, :]
        data = np.asarray(data)
        SequenceLabeller.load_sequences(data).main()

    def to_tiff(self,
                frames: slice | np.ndarray | None,
                plane: int,
                channel: int,
                output: PathLike):
        """
        Convert the selected frames to tiff file

        :param frames: selected frames. If None, convert all sequences
        :param plane: number of optical planes
        :param channel: number of PMT channel
        :param output: output filename
        """

        import tifffile

        if frames is None:
            frames = np.arange(0, self.n_frames)
        data = self.sbx_map[frames, plane, channel, :, :]
        tifffile.imwrite(output, data)


class ScanBoxViewOptions(AbstractParser):
    DESCRIPTION = 'ScanBox data view & save options'

    directory: Path = pos_argument(
        'PATH',
        validator.path.is_dir().is_exists(),
        help='directory containing .sbx/.mat scanbox output'
    )

    # -- Config ---------------
    frames: tuple[int, int] | None = argument(
        '--frames',
        type=int_tuple_type,
        default=None,
        help='indices of image sequences, if None, then all frames'
    )
    plane: int = argument('--plane', default=0, help='which optic plane')
    channel: int = argument('--channel', default=0, help='which pmt channel')

    # -- IO ---------------------
    verbose: bool = argument('--verbose', help='show meta verbose')
    show: bool = argument('--show', help='play the selected imaging sequences')
    to_tiff: Path | None = argument('--tiff', default=None, help='save sequence as tiff output')

    def run(self):
        viewer = ScanBoxView(self.directory)
        frames = np.arange(*self.frames).astype(int) if self.frames is not None else None

        if self.verbose:
            pprint(viewer.info.asdict())

        if self.show:
            viewer.show(frames, self.plane, self.channel)

        if self.to_tiff is not None:
            viewer.to_tiff(frames, self.plane, self.channel, output=self.to_tiff)


if __name__ == '__main__':
    ScanBoxViewOptions().main()
